@@ -59,6 +59,7 @@ extern "C" {
 #include <pocketsphinx_export.h>
 #include <cmdln_macro.h>
 #include <ps_lattice.h>
+#include <ps_mllr.h>
 #include <fsg_set.h>
 
 /**
@@ -188,6 +189,20 @@ fe_t *ps_get_fe(ps_decoder_t *ps);
 POCKETSPHINX_EXPORT
 feat_t *ps_get_feat(ps_decoder_t *ps);
 
+/**
+ * Adapt current acoustic model using a linear transform.
+ *
+ * @param mllr The new transform to use, or NULL to update the existing
+ *              transform.  The decoder retains ownership of this pointer,
+ *              so you should not attempt to free it manually.  Use
+ *              ps_mllr_retain() if you wish to reuse it
+ *              elsewhere.
+ * @return The updated transform object for this decoder, or
+ *         NULL on failure.
+ */
+POCKETSPHINX_EXPORT
+ps_mllr_t *ps_update_mllr(ps_decoder_t *ps, ps_mllr_t *mllr);
+
 /**
  * Get the language model set object for this decoder.
  *
@@ -246,6 +261,36 @@ fsg_set_t *ps_get_fsgset(ps_decoder_t *ps);
 POCKETSPHINX_EXPORT
 fsg_set_t *ps_update_fsgset(ps_decoder_t *ps);
 
+/**
+ * Reload the pronunciation dictionary from a file.
+ *
+ * This function replaces the current pronunciation dictionary with
+ * the one stored in dictfile.  This also causes the active search
+ * module(s) to be reinitialized, in the same manner as calling
+ * ps_add_word() with update=TRUE.
+ *
+ * @param dictfile Path to dictionary file to load.
+ * @param fdictfile Path to filler dictionary to load, or NULL to keep
+ *                  the existing filler dictionary.
+ * @param format Format of the dictionary file, or NULL to determine
+ *               automatically (currently unused,should be NULL)
+ */
+POCKETSPHINX_EXPORT
+int ps_load_dict(ps_decoder_t *ps, char const *dictfile,
+                 char const *fdictfile, char const *format);
+
+/**
+ * Dump the current pronunciation dictionary to a file.
+ *
+ * This function dumps the current pronunciation dictionary to a tex
+ *
+ * @param dictfile Path to file where dictionary will be written.
+ * @param format Format of the dictionary file, or NULL for the
+ *               default (text) format (currently unused, should be NULL)
+ */
+POCKETSPHINX_EXPORT
+int ps_save_dict(ps_decoder_t *ps, char const *dictfile, char const *format);
+
 /**
  * Add a word to the pronunciation dictionary.
  *
@@ -581,8 +626,8 @@ void ps_get_all_time(ps_decoder_t *ps, double *out_nspeech,
 /**
  * @mainpage PocketSphinx API Documentation
  * @author David Huggins-Daines <dhuggins@cs.cmu.edu>
- * @version 0.5
- * @date April, 2008
+ * @version 0.6
+ * @date March, 2010
  *
  * @section intro_sec Introduction
  *
