@@ -46,6 +46,7 @@
 /* SphinxBase headers. */
 #include <ckd_alloc.h>
 #include <listelem_alloc.h>
+#include <err.h>
 
 /* Local headers. */
 #include "pocketsphinx_internal.h"
@@ -55,9 +56,9 @@
 #include "ngram_search_fwdflat.h"
 
 static int ngram_search_start(ps_search_t *search);
-static int ngram_search_step(ps_search_t *search);
+static int ngram_search_step(ps_search_t *search, int frame_idx);
 static int ngram_search_finish(ps_search_t *search);
-static int ngram_search_reinit(ps_search_t *search);
+static int ngram_search_reinit(ps_search_t *search, dict_t *dict, dict2pid_t *d2p);
 static char const *ngram_search_hyp(ps_search_t *search, int32 *out_score);
 static int32 ngram_search_prob(ps_search_t *search);
 static ps_seg_t *ngram_search_seg_iter(ps_search_t *search, int32 *out_score);
@@ -78,7 +79,7 @@ static ps_searchfuncs_t ngram_funcs = {
 static void
 ngram_search_update_widmap(ngram_search_t *ngs)
 {
-    char const **words;
+    const char **words;
     int32 i, n_words;
 
     /* It's okay to include fillers since they won't be in the LM */
@@ -86,7 +87,7 @@ ngram_search_update_widmap(ngram_search_t *ngs)
     words = ckd_calloc(n_words, sizeof(*words));
     /* This will include alternates, again, that's okay since they aren't in the LM */
     for (i = 0; i < n_words; ++i)
-        words[i] = (const char *)dict_word_str(ps_search_dict(ngs), i);
+        words[i] = (const char *)dict_wordstr(ps_search_dict(ngs), i);
     ngram_model_set_map_words(ngs->lmset, words, n_words);
     ckd_free(words);
 }
@@ -131,21 +132,26 @@ ngram_search_calc_beams(ngram_search_t *ngs)
         / cmd_ln_float32_r(config, "-lw");
 
     /* Acoustic score scale for posterior probabilities. */
-    ngs->ascale = 1.0f / cmd_ln_float32_r(config, "-ascale");
+    ngs->ascale = 1.0 / cmd_ln_float32_r(config, "-ascale");
 }
 
 ps_search_t *
 ngram_search_init(cmd_ln_t *config,
 		  acmod_t *acmod,
-		  dict_t *dict)
+		  dict_t *dict,
+                  dict2pid_t *d2p)
 {
     ngram_search_t *ngs;
     const char *path;
 
     ngs = ckd_calloc(1, sizeof(*ngs));
-    ps_search_init(&ngs->base, &ngram_funcs, config, acmod, dict);
+    ps_search_init(&ngs->base, &ngram_funcs, config, acmod, dict, d2p);
     ngs->hmmctx = hmm_context_init(bin_mdef_n_emit_state(acmod->mdef),
                                    acmod->tmat->tp, NULL, acmod->mdef->sseq);
+    if (ngs->hmmctx == NULL) {
+        ps_search_free(ps_search_base(ngs));
+        return NULL;
+    }
     ngs->chan_alloc = listelem_alloc_init(sizeof(chan_t));
     ngs->root_chan_alloc = listelem_alloc_init(sizeof(root_chan_t));
     ngs->latnode_alloc = listelem_alloc_init(sizeof(ps_latnode_t));
@@ -154,14 +160,14 @@ ngram_search_init(cmd_ln_t *config,
     ngram_search_calc_beams(ngs);
 
     /* Allocate a billion different tables for stuff. */
-    ngs->word_chan = ckd_calloc(dict->dict_entry_count,
+    ngs->word_chan = ckd_calloc(dict_size(dict),
                                 sizeof(*ngs->word_chan));
-    ngs->word_lat_idx = ckd_calloc(dict->dict_entry_count,
+    ngs->word_lat_idx = ckd_calloc(dict_size(dict),
                                    sizeof(*ngs->word_lat_idx));
     ngs->zeroPermTab = ckd_calloc(bin_mdef_n_ciphone(acmod->mdef),
                                   sizeof(*ngs->zeroPermTab));
-    ngs->word_active = bitvec_alloc(dict->dict_entry_count);
-    ngs->last_ltrans = ckd_calloc(dict->dict_entry_count,
+    ngs->word_active = bitvec_alloc(dict_size(dict));
+    ngs->last_ltrans = ckd_calloc(dict_size(dict),
                                   sizeof(*ngs->last_ltrans));
 
     /* FIXME: All these structures need to be made dynamic with
@@ -179,7 +185,7 @@ ngram_search_init(cmd_ln_t *config,
     ++ngs->bp_table_idx; /* Make bptableidx[-1] valid */
 
     /* Allocate active word list array */
-    ngs->active_word_list = ckd_calloc_2d(2, dict->dict_entry_count,
+    ngs->active_word_list = ckd_calloc_2d(2, dict_size(dict),
                                           sizeof(**ngs->active_word_list));
 
     /* Load language model(s) */
@@ -236,15 +242,31 @@ error_out:
 }
 
 static int
-ngram_search_reinit(ps_search_t *search)
+ngram_search_reinit(ps_search_t *search, dict_t *dict, dict2pid_t *d2p)
 {
     ngram_search_t *ngs = (ngram_search_t *)search;
+    int old_n_words;
     int rv = 0;
 
-    /*
-     * NOTE!!! This is not a general-purpose reinit function.  It only
-     * deals with updates to the language model set and beam widths.
-     */
+    /* Update the number of words. */
+    old_n_words = search->n_words;
+    if (old_n_words != dict_size(dict)) {
+        search->n_words = dict_size(dict);
+        /* Reallocate these temporary arrays. */
+        ckd_free(ngs->word_lat_idx);
+        ckd_free(ngs->word_active);
+        ckd_free(ngs->last_ltrans);
+        ckd_free_2d(ngs->active_word_list);
+        ngs->word_lat_idx = ckd_calloc(search->n_words, sizeof(*ngs->word_lat_idx));
+        ngs->word_active = bitvec_alloc(search->n_words);
+        ngs->last_ltrans = ckd_calloc(search->n_words, sizeof(*ngs->last_ltrans));
+        ngs->active_word_list
+            = ckd_calloc_2d(2, search->n_words,
+                            sizeof(**ngs->active_word_list));
+    }
+
+    /* Free old dict2pid, dict */
+    ps_search_base_reinit(search, dict, d2p);
 
     /* Update beam widths. */
     ngram_search_calc_beams(ngs);
@@ -252,7 +274,7 @@ ngram_search_reinit(ps_search_t *search)
     /* Update word mappings. */
     ngram_search_update_widmap(ngs);
 
-    /* Now rebuild lextrees or what have you. */
+    /* Now rebuild lextrees. */
     if (ngs->fwdtree) {
         if ((rv = ngram_fwdtree_reinit(ngs)) < 0)
             return rv;
@@ -288,7 +310,8 @@ ngram_search_free(ps_search_t *search)
     bitvec_free(ngs->word_active);
     ckd_free(ngs->bp_table);
     ckd_free(ngs->bscore_stack);
-    ckd_free(ngs->bp_table_idx - 1);
+    if (ngs->bp_table_idx != NULL)
+        ckd_free(ngs->bp_table_idx - 1);
     ckd_free_2d(ngs->active_word_list);
     ckd_free(ngs->last_ltrans);
     ckd_free(ngs);
@@ -320,20 +343,25 @@ static void
 cache_bptable_paths(ngram_search_t *ngs, int32 bp)
 {
     int32 w, prev_bp;
-    bptbl_t *bpe;
+    bptbl_t *be;
 
-    bpe = &(ngs->bp_table[bp]);
+    assert(bp != NO_BP);
+
+    be = &(ngs->bp_table[bp]);
     prev_bp = bp;
-    w = bpe->wid;
-    /* FIXME: This isn't the ideal way to tell if it's a filler. */
-    while (w >= ps_search_silence_wid(ngs)) {
+    w = be->wid;
+
+    while (dict_filler_word(ps_search_dict(ngs), w)) {
         prev_bp = ngs->bp_table[prev_bp].bp;
+        if (prev_bp == NO_BP)
+            return;
         w = ngs->bp_table[prev_bp].wid;
     }
-    bpe->real_wid = dict_base_wid(ps_search_dict(ngs), w);
+
+    be->real_wid = dict_basewid(ps_search_dict(ngs), w);
 
     prev_bp = ngs->bp_table[prev_bp].bp;
-    bpe->prev_real_wid =
+    be->prev_real_wid =
         (prev_bp != NO_BP) ? ngs->bp_table[prev_bp].real_wid : -1;
 }
 
@@ -343,21 +371,32 @@ ngram_search_save_bp(ngram_search_t *ngs, int frame_idx,
 {
     int32 _bp_;
 
+    /* Look for an existing exit for this word in this frame. */
     _bp_ = ngs->word_lat_idx[w];
     if (_bp_ != NO_BP) {
-        if (ngs->bp_table[_bp_].score < score) {
+        /* Keep only the best scoring one (this is a potential source
+         * of search errors...) */
+        if (ngs->bp_table[_bp_].score WORSE_THAN score) {
             if (ngs->bp_table[_bp_].bp != path) {
                 ngs->bp_table[_bp_].bp = path;
                 cache_bptable_paths(ngs, _bp_);
             }
             ngs->bp_table[_bp_].score = score;
         }
+        /* But do keep track of scores for all right contexts, since
+         * we need them to determine the starting path scores for any
+         * successors of this word exit. */
         ngs->bscore_stack[ngs->bp_table[_bp_].s_idx + rc] = score;
     }
     else {
         int32 i, rcsize, *bss;
-        dict_entry_t *de;
-        bptbl_t *bpe;
+        bptbl_t *be;
+
+        /* This might happen if recognition fails. */
+        if (ngs->bpidx == NO_BP) {
+            E_ERROR("No entries in backpointer table!");
+            return;
+        }
 
         /* Expand the backpointer tables if necessary. */
         if (ngs->bpidx >= ngs->bp_table_size) {
@@ -376,24 +415,28 @@ ngram_search_save_bp(ngram_search_t *ngs, int frame_idx,
             E_INFO("Resized score stack to %d entries\n", ngs->bscore_stack_size);
         }
 
-        de = ps_search_dict(ngs)->dict_list[w];
         ngs->word_lat_idx[w] = ngs->bpidx;
-        bpe = &(ngs->bp_table[ngs->bpidx]);
-        bpe->wid = w;
-        bpe->frame = frame_idx;
-        bpe->bp = path;
-        bpe->score = score;
-        bpe->s_idx = ngs->bss_head;
-        bpe->valid = TRUE;
+        be = &(ngs->bp_table[ngs->bpidx]);
+        be->wid = w;
+        be->frame = frame_idx;
+        be->bp = path;
+        be->score = score;
+        be->s_idx = ngs->bss_head;
+        be->valid = TRUE;
 
-        if ((de->len != 1) && (de->mpx)) {
-            bpe->r_diph = de->phone_ids[de->len - 1];
-            rcsize = ps_search_dict(ngs)->rcFwdSizeTable[bpe->r_diph];
-        }
-        else {
-            bpe->r_diph = -1;
+        /* DICT2PID */
+        /* Get diphone ID for final phone and number of ssids corresponding to it. */
+        be->last_phone = dict_last_phone(ps_search_dict(ngs),w);
+        if (dict_is_single_phone(ps_search_dict(ngs), w)) {
+            be->last2_phone = -1;
             rcsize = 1;
         }
+        else {
+            be->last2_phone = dict_second_last_phone(ps_search_dict(ngs),w);
+            rcsize = dict2pid_rssid(ps_search_dict2pid(ngs),
+                                    be->last_phone, be->last2_phone)->n_ssid;
+        }
+        /* Allocate some space on the bscore_stack for all of these triphones. */
         for (i = rcsize, bss = ngs->bscore_stack + ngs->bss_head; i > 0; --i, bss++)
             *bss = WORST_SCORE;
         ngs->bscore_stack[ngs->bss_head + rc] = score;
@@ -420,7 +463,6 @@ ngram_search_find_exit(ngram_search_t *ngs, int frame_idx, int32 *out_best_score
         frame_idx = ngs->n_frame - 1;
     end_bpidx = ngs->bp_table_idx[frame_idx];
 
-    /* FIXME: WORST_SCORE has to go away and be replaced with a log-zero number. */
     best_score = WORST_SCORE;
     best_exit = NO_BP;
 
@@ -435,7 +477,7 @@ ngram_search_find_exit(ngram_search_t *ngs, int frame_idx, int32 *out_best_score
     assert(end_bpidx < ngs->bp_table_size);
     for (bp = ngs->bp_table_idx[frame_idx]; bp < end_bpidx; ++bp) {
         if (ngs->bp_table[bp].wid == ps_search_finish_wid(ngs)
-            || ngs->bp_table[bp].score > best_score) {
+            || ngs->bp_table[bp].score BETTER_THAN best_score) {
             best_score = ngs->bp_table[bp].score;
             best_exit = bp;
         }
@@ -463,12 +505,17 @@ ngram_search_bp_hyp(ngram_search_t *ngs, int bpidx)
     while (bp != NO_BP) {
         bptbl_t *be = &ngs->bp_table[bp];
         bp = be->bp;
-        if (ISA_REAL_WORD(ngs, be->wid))
-            len += strlen(dict_base_str(ps_search_dict(ngs), be->wid)) + 1;
+        if (dict_real_word(ps_search_dict(ngs), be->wid))
+            len += strlen(dict_basestr(ps_search_dict(ngs), be->wid)) + 1;
     }
 
     ckd_free(base->hyp_str);
+    if (len == 0) {
+	base->hyp_str = NULL;
+	return base->hyp_str;
+    }
     base->hyp_str = ckd_calloc(1, len);
+
     bp = bpidx;
     c = base->hyp_str + len - 1;
     while (bp != NO_BP) {
@@ -476,10 +523,10 @@ ngram_search_bp_hyp(ngram_search_t *ngs, int bpidx)
         size_t len;
 
         bp = be->bp;
-        if (ISA_REAL_WORD(ngs, be->wid)) {
-            len = strlen(dict_base_str(ps_search_dict(ngs), be->wid));
+        if (dict_real_word(ps_search_dict(ngs), be->wid)) {
+            len = strlen(dict_basestr(ps_search_dict(ngs), be->wid));
             c -= len;
-            memcpy(c, dict_base_str(ps_search_dict(ngs), be->wid), len);
+            memcpy(c, dict_basestr(ps_search_dict(ngs), be->wid), len);
             if (c > base->hyp_str) {
                 --c;
                 *c = ' ';
@@ -493,37 +540,44 @@ ngram_search_bp_hyp(ngram_search_t *ngs, int bpidx)
 void
 ngram_search_alloc_all_rc(ngram_search_t *ngs, int32 w)
 {
-    dict_entry_t *de;
     chan_t *hmm, *thmm;
-    uint16 *sseq_rc;             /* list of sseqid for all possible right context for w */
+    xwdssid_t *rssid;
     int32 i;
 
-    de = ps_search_dict(ngs)->dict_list[w];
-
-    assert(de->mpx);
-
-    sseq_rc = ps_search_dict(ngs)->rcFwdTable[de->phone_ids[de->len - 1]];
-
+    /* DICT2PID */
+    /* Get pointer to array of triphones for final diphone. */
+    assert(!dict_is_single_phone(ps_search_dict(ngs), w));
+    rssid = dict2pid_rssid(ps_search_dict2pid(ngs),
+                           dict_last_phone(ps_search_dict(ngs),w),
+                           dict_second_last_phone(ps_search_dict(ngs),w));
     hmm = ngs->word_chan[w];
-    if ((hmm == NULL) || (hmm->hmm.s.ssid != *sseq_rc)) {
+    if ((hmm == NULL) || (hmm_nonmpx_ssid(&hmm->hmm) != rssid->ssid[0])) {
         hmm = listelem_malloc(ngs->chan_alloc);
         hmm->next = ngs->word_chan[w];
         ngs->word_chan[w] = hmm;
 
         hmm->info.rc_id = 0;
-        hmm->ciphone = de->ci_phone_ids[de->len - 1];
-        hmm_init(ngs->hmmctx, &hmm->hmm, FALSE, *sseq_rc, hmm->ciphone);
+        hmm->ciphone = dict_last_phone(ps_search_dict(ngs),w);
+        hmm_init(ngs->hmmctx, &hmm->hmm, FALSE, rssid->ssid[0], hmm->ciphone);
+        E_DEBUG(3,("allocated rc_id 0 ssid %d ciphone %d lc %d word %s\n",
+                   rssid->ssid[0], hmm->ciphone,
+                   dict_second_last_phone(ps_search_dict(ngs),w),
+                   dict_wordstr(ps_search_dict(ngs),w)));
     }
-    for (i = 1, sseq_rc++; *sseq_rc != 65535; sseq_rc++, i++) {
-        if ((hmm->next == NULL) || (hmm->next->hmm.s.ssid != *sseq_rc)) {
+    for (i = 1; i < rssid->n_ssid; ++i) {
+        if ((hmm->next == NULL) || (hmm_nonmpx_ssid(&hmm->next->hmm) != rssid->ssid[i])) {
             thmm = listelem_malloc(ngs->chan_alloc);
             thmm->next = hmm->next;
             hmm->next = thmm;
             hmm = thmm;
 
             hmm->info.rc_id = i;
-            hmm->ciphone = de->ci_phone_ids[de->len - 1];
-            hmm_init(ngs->hmmctx, &hmm->hmm, FALSE, *sseq_rc, hmm->ciphone);
+            hmm->ciphone = dict_last_phone(ps_search_dict(ngs),w);
+            hmm_init(ngs->hmmctx, &hmm->hmm, FALSE, rssid->ssid[i], hmm->ciphone);
+            E_DEBUG(3,("allocated rc_id %d ssid %d ciphone %d lc %d word %s\n",
+                       i, rssid->ssid[i], hmm->ciphone,
+                       dict_second_last_phone(ps_search_dict(ngs),w),
+                       dict_wordstr(ps_search_dict(ngs),w)));
         }
         else
             hmm = hmm->next;
@@ -543,61 +597,74 @@ ngram_search_free_all_rc(ngram_search_t *ngs, int32 w)
     ngs->word_chan[w] = NULL;
 }
 
+int32
+ngram_search_exit_score(ngram_search_t *ngs, bptbl_t *pbe, int rcphone)
+{
+    /* DICT2PID */
+    /* Get the mapping from right context phone ID to index in the
+     * right context table and the bscore_stack. */
+    if (pbe->last2_phone == -1) {
+        /* No right context for single phone predecessor words. */
+        return ngs->bscore_stack[pbe->s_idx];
+    }
+    else {
+        xwdssid_t *rssid;
+        /* Find the index for the last diphone of the previous word +
+         * the first phone of the current word. */
+        rssid = dict2pid_rssid(ps_search_dict2pid(ngs),
+                               pbe->last_phone, pbe->last2_phone);
+        return ngs->bscore_stack[pbe->s_idx + rssid->cimap[rcphone]];
+    }
+}
+
 /*
- * Compute acoustic and LM scores for each BPTable entry (segment).
+ * Compute acoustic and LM scores for a BPTable entry (segment).
  */
 void
-ngram_compute_seg_scores(ngram_search_t *ngs, float32 lwf)
+ngram_compute_seg_score(ngram_search_t *ngs, bptbl_t *be, float32 lwf,
+                        int32 *out_ascr, int32 *out_lscr)
 {
-    int32 bp, start_score;
-    bptbl_t *bpe, *p_bpe;
-    uint16 *rcpermtab;
-    dict_entry_t *de;
+    bptbl_t *pbe;
+    int32 start_score;
 
-    for (bp = 0; bp < ngs->bpidx; bp++) {
-        bpe = ngs->bp_table + bp;
-
-        /* Start of utterance. */
-        if (bpe->bp == NO_BP) {
-            bpe->ascr = bpe->score;
-            bpe->lscr = 0;
-            continue;
-        }
-
-        /* Otherwise, calculate lscr and ascr. */
-        de = ps_search_dict(ngs)->dict_list[bpe->wid];
-        p_bpe = ngs->bp_table + bpe->bp;
-        rcpermtab = (p_bpe->r_diph >= 0) ?
-            ps_search_dict(ngs)->rcFwdPermTable[p_bpe->r_diph] : ngs->zeroPermTab;
-        start_score =
-            ngs->bscore_stack[p_bpe->s_idx + rcpermtab[de->ci_phone_ids[0]]];
-        /* FIXME: WORST_SCORE shouldn't propagate, but sometimes it
-           does.  We cannot allow it to go any further because that
-           will result in positive acoustic scores.  Tracing the
-           source of this may be a bit involved. */
-        if (start_score == WORST_SCORE)
-            start_score = 0;
-
-        /* FIXME: These result in positive acoustic scores when filler
-           words have non-filler pronunciations.  That whole business
-           is still pretty much broken but at least it doesn't
-           segfault. */
-        if (bpe->wid == ps_search_silence_wid(ngs)) {
-            bpe->lscr = ngs->silpen;
-        }
-        else if (ISA_FILLER_WORD(ngs, bpe->wid)) {
-            bpe->lscr = ngs->fillpen;
-        }
-        else {
-            int32 n_used;
-            bpe->lscr = ngram_tg_score(ngs->lmset, de->wid,
-                                       p_bpe->real_wid,
-                                       p_bpe->prev_real_wid, &n_used);
-			/* FIXME: Floating-point conversion = SLOW */
-            bpe->lscr = (int32)(bpe->lscr * lwf);
-        }
-        bpe->ascr = bpe->score - start_score - bpe->lscr;
+    /* Start of utterance. */
+    if (be->bp == NO_BP) {
+        *out_ascr = be->score;
+        *out_lscr = 0;
+        return;
     }
+
+    /* Otherwise, calculate lscr and ascr. */
+    pbe = ngs->bp_table + be->bp;
+    start_score = ngram_search_exit_score(ngs, pbe,
+                                 dict_first_phone(ps_search_dict(ngs),be->wid));
+
+    /* FIXME: WORST_SCORE shouldn't propagate, but sometimes it
+       does.  We cannot allow it to go any further because that
+       will result in positive acoustic scores.  Tracing the
+       source of this may be a bit involved. */
+    if (start_score == WORST_SCORE)
+        start_score = 0;
+
+    /* FIXME: These result in positive acoustic scores when filler
+       words have non-filler pronunciations.  That whole business
+       is still pretty much broken but at least it doesn't
+       segfault. */
+    if (be->wid == ps_search_silence_wid(ngs)) {
+        *out_lscr = ngs->silpen;
+    }
+    else if (dict_filler_word(ps_search_dict(ngs), be->wid)) {
+        *out_lscr = ngs->fillpen;
+    }
+    else {
+        int32 n_used;
+        *out_lscr = ngram_tg_score(ngs->lmset,
+                                   be->real_wid,
+                                   pbe->real_wid,
+                                   pbe->prev_real_wid, &n_used);
+        *out_lscr = *out_lscr * lwf;
+    }
+    *out_ascr = be->score - start_score - *out_lscr;
 }
 
 static int
@@ -617,18 +684,35 @@ ngram_search_start(ps_search_t *search)
 }
 
 static int
-ngram_search_step(ps_search_t *search)
+ngram_search_step(ps_search_t *search, int frame_idx)
 {
     ngram_search_t *ngs = (ngram_search_t *)search;
 
     if (ngs->fwdtree)
-        return ngram_fwdtree_search(ngs);
+        return ngram_fwdtree_search(ngs, frame_idx);
     else if (ngs->fwdflat)
-        return ngram_fwdflat_search(ngs);
+        return ngram_fwdflat_search(ngs, frame_idx);
     else
         return -1;
 }
 
+static void
+dump_bptable(ngram_search_t *ngs)
+{
+    int i;
+    E_INFO("Backpointer table (%d entries):\n", ngs->bpidx);
+    for (i = 0; i < ngs->bpidx; ++i) {
+        E_INFO_NOFN("%-5d %-10s start %-3d end %-3d score %-8d bp\n", /* %-3d history %08x\n", */
+                    i, dict_wordstr(ps_search_dict(ngs), ngs->bp_table[i].wid),
+                    ngs->bp_table[i].bp == -1 ? 0 : 
+                    ngs->bp_table[ngs->bp_table[i].bp].frame + 1,
+                    ngs->bp_table[i].frame,
+                    ngs->bp_table[i].score,
+                    ngs->bp_table[i].bp);
+        /* ngs->bp_table[i].hist_hash); */
+    }
+}
+
 static int
 ngram_search_finish(ps_search_t *search)
 {
@@ -636,22 +720,27 @@ ngram_search_finish(ps_search_t *search)
 
     if (ngs->fwdtree) {
         ngram_fwdtree_finish(ngs);
+        /* dump_bptable(ngs); */
 
         /* Now do fwdflat search in its entirety, if requested. */
         if (ngs->fwdflat) {
-            int nfr;
-
+            int i;
             /* Rewind the acoustic model. */
-            acmod_rewind(ps_search_acmod(ngs));
+            if (acmod_rewind(ps_search_acmod(ngs)) < 0)
+                return -1;
             /* Now redo search. */
             ngram_fwdflat_start(ngs);
-            while ((nfr = ngram_fwdflat_search(ngs)) > 0) {
-                /* Do nothing! */
+            i = 0;
+            while (ps_search_acmod(ngs)->n_feat_frame > 0) {
+                int nfr;
+                if ((nfr = ngram_fwdflat_search(ngs, i)) < 0)
+                    return nfr;
+                acmod_advance(ps_search_acmod(ngs));
+                ++i;
             }
-            if (nfr < 0)
-                return nfr;
             ngram_fwdflat_finish(ngs);
             /* And now, we should have a result... */
+            /* dump_bptable(ngs); */
         }
     }
     else if (ngs->fwdflat) {
@@ -721,42 +810,36 @@ ngram_search_bp2itor(ps_seg_t *seg, int bp)
 
     be = &ngs->bp_table[bp];
     pbe = be->bp == -1 ? NULL : &ngs->bp_table[be->bp];
-    seg->word = dict_word_str(ps_search_dict(ngs), be->wid);
+    seg->word = dict_wordstr(ps_search_dict(ngs), be->wid);
     seg->ef = be->frame;
     seg->sf = pbe ? pbe->frame + 1 : 0;
     seg->prob = 0; /* Bogus value... */
     /* Compute acoustic and LM scores for this segment. */
     if (pbe == NULL) {
-        seg->ascr = be->ascr = be->score;
-        seg->lscr = be->lscr = 0;
+        seg->ascr = be->score;
+        seg->lscr = 0;
         seg->lback = 0;
     }
     else {
-        dict_entry_t *de;
-        uint16 *rcpermtab;
         int32 start_score;
 
-        de = seg->search->dict->dict_list[be->wid];
         /* Find ending path score of previous word. */
-        rcpermtab = (pbe->r_diph >= 0)
-            ? seg->search->dict->rcFwdPermTable[pbe->r_diph]
-            : ngs->zeroPermTab;
-        start_score = ngs->bscore_stack[pbe->s_idx + rcpermtab[de->ci_phone_ids[0]]];
+        start_score = ngram_search_exit_score(ngs, pbe,
+                                     dict_first_phone(ps_search_dict(ngs), be->wid));
         if (be->wid == ps_search_silence_wid(ngs)) {
-            be->lscr = ngs->silpen;
+            seg->lscr = ngs->silpen;
         }
-        else if (ISA_FILLER_WORD(ngs, be->wid)) {
-            be->lscr = ngs->fillpen;
+        else if (dict_filler_word(ps_search_dict(ngs), be->wid)) {
+            seg->lscr = ngs->fillpen;
         }
         else {
-            be->lscr = ngram_tg_score(ngs->lmset, de->wid,
-                                      pbe->real_wid,
-                                      pbe->prev_real_wid, &seg->lback);
-            be->lscr = (int32)(be->lscr * seg->lwf);
+            seg->lscr = ngram_tg_score(ngs->lmset,
+                                       be->real_wid,
+                                       pbe->real_wid,
+                                       pbe->prev_real_wid, &seg->lback);
+            seg->lscr = (int32)(seg->lscr * seg->lwf);
         }
-        be->ascr = be->score - start_score - be->lscr;
-        seg->ascr = be->ascr;
-        seg->lscr = be->lscr;
+        seg->ascr = be->score - start_score - seg->lscr;
     }
 }
 
@@ -854,7 +937,7 @@ ngram_search_seg_iter(ps_search_t *search, int32 *out_score)
         return ngram_search_bp_iter(ngs, bpidx,
                                     /* but different language weights... */
                                     (ngs->done && ngs->fwdflat)
-                                    ? ngs->fwdflat_fwdtree_lw_ratio : 1.0f);
+                                    ? ngs->fwdflat_fwdtree_lw_ratio : 1.0);
     }
 
     return NULL;
@@ -892,6 +975,7 @@ create_dag_nodes(ngram_search_t *ngs, ps_lattice_t *dag)
         int32 sf, ef, wid;
         ps_latnode_t *node;
 
+        /* Skip invalid backpointers (these result from -maxwpf pruning) */
         if (!bp_ptr->valid)
             continue;
 
@@ -905,9 +989,9 @@ create_dag_nodes(ngram_search_t *ngs, ps_lattice_t *dag)
             continue;
 
         /* Skip if word not in LM */
-        if ((!ISA_FILLER_WORD(ngs, wid))
+        if ((!dict_filler_word(ps_search_dict(ngs), wid))
             && (!ngram_model_set_known_wid(ngs->lmset,
-                                           dict_base_wid(ps_search_dict(ngs), wid))))
+                                           dict_basewid(ps_search_dict(ngs), wid))))
             continue;
 
         /* See if bptbl entry <wid,sf> already in lattice */
@@ -989,16 +1073,19 @@ find_end_node(ngram_search_t *ngs, ps_lattice_t *dag, float32 lwf)
                                ngs->bp_table[bp].real_wid,
                                ngs->bp_table[bp].prev_real_wid,
                                &n_used);
-		/* FIXME: Floating-point conversion = SLOW */
-        l_scr = (int32)(l_scr * lwf);
+        l_scr = l_scr * lwf;
 
-        if (ngs->bp_table[bp].score + l_scr > bestscore) {
+        if (ngs->bp_table[bp].score + l_scr BETTER_THAN bestscore) {
             bestscore = ngs->bp_table[bp].score + l_scr;
             bestbp = bp;
         }
     }
+    if (bestbp == NO_BP) {
+        E_ERROR("No word exits found in last frame, assuming no recognition\n");
+        return NULL;
+    }
     E_WARN("</s> not found in last frame, using %s instead\n",
-           dict_base_str(ps_search_dict(ngs), ngs->bp_table[bestbp].wid));
+           dict_basestr(ps_search_dict(ngs), ngs->bp_table[bestbp].wid));
 
     /* Now find the node that corresponds to it. */
     for (node = dag->nodes; node; node = node->next) {
@@ -1008,7 +1095,7 @@ find_end_node(ngram_search_t *ngs, ps_lattice_t *dag, float32 lwf)
 
     /* FIXME: This seems to happen a lot! */
     E_ERROR("Failed to find DAG node corresponding to %s\n",
-           dict_base_str(ps_search_dict(ngs), ngs->bp_table[bestbp].wid));
+           dict_basestr(ps_search_dict(ngs), ngs->bp_table[bestbp].wid));
     return NULL;
 }
 
@@ -1018,13 +1105,19 @@ find_end_node(ngram_search_t *ngs, ps_lattice_t *dag, float32 lwf)
 ps_lattice_t *
 ngram_search_lattice(ps_search_t *search)
 {
-    int32 i, ef, lef, score, bss_offset;
+    int32 i, ef, lef, score, ascr, lscr;
     ps_latnode_t *node, *from, *to;
     ngram_search_t *ngs;
     ps_lattice_t *dag;
+    float lwf;
 
     ngs = (ngram_search_t *)search;
 
+    /* If the best score is WORST_SCORE or worse, there is no way to
+     * make a lattice. */
+    if (ngs->best_score == WORST_SCORE || ngs->best_score WORSE_THAN WORST_SCORE)
+        return NULL;
+
     /* Check to see if a lattice has previously been created over the
      * same number of frames, and reuse it if so. */
     if (search->dag && search->dag->n_frames == ngs->n_frame)
@@ -1035,19 +1128,18 @@ ngram_search_lattice(ps_search_t *search)
     search->dag = NULL;
     dag = ps_lattice_init_search(search, ngs->n_frame);
     /* Compute these such that they agree with the fwdtree language weight. */
-    ngram_compute_seg_scores(ngs,
-                             ngs->fwdflat
-                             ? ngs->fwdflat_fwdtree_lw_ratio : 1.0);
+    lwf = ngs->fwdflat ? ngs->fwdflat_fwdtree_lw_ratio : 1.0;
     create_dag_nodes(ngs, dag);
     if ((dag->start = find_start_node(ngs, dag)) == NULL)
         goto error_out;
     if ((dag->end = find_end_node(ngs, dag, ngs->bestpath_fwdtree_lw_ratio)) == NULL)
         goto error_out;
     E_INFO("lattice start node %s.%d end node %s.%d\n",
-           dict_word_str(search->dict, dag->start->wid), dag->start->sf,
-           dict_word_str(search->dict, dag->end->wid), dag->end->sf);
+           dict_wordstr(search->dict, dag->start->wid), dag->start->sf,
+           dict_wordstr(search->dict, dag->end->wid), dag->end->sf);
 
-    dag->final_node_ascr = ngs->bp_table[dag->end->lef].ascr;
+    ngram_compute_seg_score(ngs, ngs->bp_table + dag->end->lef, lwf,
+                            &dag->final_node_ascr, &lscr);
 
     /*
      * At this point, dag->nodes is ordered such that nodes earlier in
@@ -1086,16 +1178,13 @@ ngram_search_lattice(ps_search_t *search)
                 continue;
 
             /* Find acoustic score from.sf->to.sf-1 with right context = to */
-            if (bp_ptr->r_diph >= 0)
-                bss_offset =
-                    search->dict->rcFwdPermTable[bp_ptr->r_diph]
-                    [search->dict->dict_list[to->wid]->ci_phone_ids[0]];
-            else
-                bss_offset = 0;
-            score =
-                (ngs->bscore_stack[bp_ptr->s_idx + bss_offset] - bp_ptr->score) +
-                bp_ptr->ascr;
-            if (score > 0) {
+            ngram_compute_seg_score(ngs, bp_ptr, lwf,
+                                    &ascr, &lscr);
+            /* Remove context score calculated above (FIXME: just don't calculate it...) */
+            score = (ngram_search_exit_score(ngs, bp_ptr,
+                                    dict_first_phone(ps_search_dict(ngs), to->wid))
+                     - bp_ptr->score + ascr);
+            if (score BETTER_THAN 0) {
                 /* Scores must be negative, or Bad Things will happen.
                    In general, they are, except in corner cases
                    involving filler words.  We don't want to throw any
@@ -1104,7 +1193,7 @@ ngram_search_lattice(ps_search_t *search)
                 ps_lattice_link(dag, from, to, -424242, bp_ptr->frame);
                 from->reachable = TRUE;
             }
-            else if (score > WORST_SCORE) {
+            else if (score BETTER_THAN WORST_SCORE) {
                 ps_lattice_link(dag, from, to, score, bp_ptr->frame);
                 from->reachable = TRUE;
             }
@@ -1122,13 +1211,26 @@ ngram_search_lattice(ps_search_t *search)
         node->fef = ngs->bp_table[node->fef].frame;
         node->lef = ngs->bp_table[node->lef].frame;
         /* Find base wid for nodes. */
-        node->basewid = dict_base_wid(search->dict, node->wid);
+        node->basewid = dict_basewid(search->dict, node->wid);
+    }
+
+    /* Link nodes with alternate pronunciations at the same timepoint. */
+    for (node = dag->nodes; node; node = node->next) {
+        ps_latnode_t *alt;
+        /* Scan forward to find the next alternate, then stop. */
+        for (alt = node->next; alt && alt->sf == node->sf; alt = alt->next) {
+            if (alt->basewid == node->basewid) {
+                alt->alt = node->alt;
+                node->alt = alt;
+                break;
+            }
+        }
     }
 
     /* Minor hack: If the final node is a filler word and not </s>,
      * then set its base word ID to </s>, so that the language model
      * scores won't be screwed up. */
-    if (ISA_FILLER_WORD(ngs, dag->end->wid))
+    if (dict_filler_word(ps_search_dict(ngs), dag->end->wid))
         dag->end->basewid = ps_search_finish_wid(ngs);
 
     /* Free nodes unreachable from dag->end and their links */
