@@ -49,6 +49,7 @@
 #include "cmdln_macro.h"
 #include "pocketsphinx_internal.h"
 #include "ps_lattice_internal.h"
+#include "phone_loop_search.h"
 #include "fsg_search_internal.h"
 #include "ngram_search.h"
 #include "ngram_search_fwdtree.h"
@@ -70,6 +71,18 @@ file_exists(const char *path)
     return (tmp != NULL);
 }
 
+static int
+hmmdir_exists(const char *path)
+{
+    FILE *tmp;
+    char *mdef = string_join(path, "/mdef", NULL);
+
+    tmp = fopen(mdef, "rb");
+    if (tmp) fclose(tmp);
+    ckd_free(mdef);
+    return (tmp != NULL);
+}
+
 static void
 ps_add_file(ps_decoder_t *ps, const char *arg,
             const char *hmmdir, const char *file)
@@ -84,13 +97,52 @@ ps_add_file(ps_decoder_t *ps, const char *arg,
 static void
 ps_init_defaults(ps_decoder_t *ps)
 {
-    char const *hmmdir;
+    char const *hmmdir, *lmfile, *dictfile;
 
     /* Disable memory mapping on Blackfin (FIXME: should be uClinux in general). */
 #ifdef __ADSPBLACKFIN__
     E_INFO("Will not use mmap() on uClinux/Blackfin.");
     cmd_ln_set_boolean_r(ps->config, "-mmap", FALSE);
 #endif
+
+#ifdef MODELDIR
+    /* Set default acoustic and language models. */
+    hmmdir = cmd_ln_str_r(ps->config, "-hmm");
+    lmfile = cmd_ln_str_r(ps->config, "-lm");
+    dictfile = cmd_ln_str_r(ps->config, "-dict");
+    if (hmmdir == NULL && hmmdir_exists(MODELDIR "/hmm/en_US/hub4wsj_sc_8k")) {
+        hmmdir = MODELDIR "/hmm/en_US/hub4wsj_sc_8k";
+        cmd_ln_set_str_r(ps->config, "-hmm", hmmdir);
+    }
+    if (lmfile == NULL && !cmd_ln_str_r(ps->config, "-fsg")
+        && !cmd_ln_str_r(ps->config, "-jsgf")
+        && file_exists(MODELDIR "/lm/en_US/hub4.5000.DMP")) {
+        lmfile = MODELDIR "/lm/en_US/hub4.5000.DMP";
+        cmd_ln_set_str_r(ps->config, "-lm", lmfile);
+    }
+    if (dictfile == NULL && file_exists(MODELDIR "/lm/en_US/cmu07a.dic")) {
+        dictfile = MODELDIR "/lm/en_US/cmu07a.dic";
+        cmd_ln_set_str_r(ps->config, "-dict", dictfile);
+    }
+
+    /* Expand acoustic and language model filenames relative to installation path. */
+    if (hmmdir && !path_is_absolute(hmmdir) && !hmmdir_exists(hmmdir)) {
+        char *tmphmm = string_join(MODELDIR "/hmm/", hmmdir, NULL);
+        cmd_ln_set_str_r(ps->config, "-hmm", tmphmm);
+        ckd_free(tmphmm);
+    }
+    if (lmfile && !path_is_absolute(lmfile) && !file_exists(lmfile)) {
+        char *tmplm = string_join(MODELDIR "/lm/", lmfile, NULL);
+        cmd_ln_set_str_r(ps->config, "-lm", tmplm);
+        ckd_free(tmplm);
+    }
+    if (dictfile && !path_is_absolute(dictfile) && !file_exists(dictfile)) {
+        char *tmpdict = string_join(MODELDIR "/lm/", dictfile, NULL);
+        cmd_ln_set_str_r(ps->config, "-dict", tmpdict);
+        ckd_free(tmpdict);
+    }
+#endif
+
     /* Get acoustic model filenames and add them to the command-line */
     if ((hmmdir = cmd_ln_str_r(ps->config, "-hmm")) != NULL) {
         ps_add_file(ps, "-mdef", hmmdir, "mdef");
@@ -99,26 +151,55 @@ ps_init_defaults(ps_decoder_t *ps)
         ps_add_file(ps, "-tmat", hmmdir, "transition_matrices");
         ps_add_file(ps, "-mixw", hmmdir, "mixture_weights");
         ps_add_file(ps, "-sendump", hmmdir, "sendump");
-        ps_add_file(ps, "-kdtree", hmmdir, "kdtrees");
         ps_add_file(ps, "-fdict", hmmdir, "noisedict");
         ps_add_file(ps, "-lda", hmmdir, "feature_transform");
         ps_add_file(ps, "-featparams", hmmdir, "feat.params");
+        ps_add_file(ps, "-senmgau", hmmdir, "senmgau");
     }
 }
 
+static void
+ps_free_searches(ps_decoder_t *ps)
+{
+    gnode_t *gn;
+
+    if (ps->searches == NULL)
+        return;
+
+    for (gn = ps->searches; gn; gn = gnode_next(gn))
+        ps_search_free(gnode_ptr(gn));
+    glist_free(ps->searches);
+    ps->searches = NULL;
+    ps->search = NULL;
+}
+
+static ps_search_t *
+ps_find_search(ps_decoder_t *ps, char const *name)
+{
+    gnode_t *gn;
+
+    for (gn = ps->searches; gn; gn = gnode_next(gn)) {
+        if (0 == strcmp(ps_search_name(gnode_ptr(gn)), name))
+            return (ps_search_t *)gnode_ptr(gn);
+    }
+    return NULL;
+}
+
 int
 ps_reinit(ps_decoder_t *ps, cmd_ln_t *config)
 {
     char const *lmfile, *lmctl = NULL;
-    gnode_t *gn;
 
     if (config && config != ps->config) {
         cmd_ln_free_r(ps->config);
         ps->config = config;
     }
+#ifndef _WIN32_WCE
     /* Set up logging. */
     if (cmd_ln_str_r(ps->config, "-logfn"))
         err_set_logfile(cmd_ln_str_r(ps->config, "-logfn"));
+#endif
+    err_set_debug_level(cmd_ln_int32_r(ps->config, "-debug"));
     ps->mfclogdir = cmd_ln_str_r(ps->config, "-mfclogdir");
     ps->rawlogdir = cmd_ln_str_r(ps->config, "-rawlogdir");
 
@@ -126,23 +207,16 @@ ps_reinit(ps_decoder_t *ps, cmd_ln_t *config)
     ps_init_defaults(ps);
 
     /* Free old searches (do this before other reinit) */
-    if (ps->searches) {
-        for (gn = ps->searches; gn; gn = gnode_next(gn))
-            ps_search_free(gnode_ptr(gn));
-        glist_free(ps->searches);
-        ps->searches = NULL;
-        ps->search = NULL;
-    }
+    ps_free_searches(ps);
+
     /* Free old acmod. */
-    if (ps->acmod) {
-        acmod_free(ps->acmod);
-        ps->acmod = NULL;
-    }
+    acmod_free(ps->acmod);
+    ps->acmod = NULL;
+
     /* Free old dictionary (must be done after the two things above) */
-    if (ps->dict) {
-        dict_free(ps->dict);
-        ps->dict = NULL;
-    }
+    dict_free(ps->dict);
+    ps->dict = NULL;
+
 
     /* Logmath computation (used in acmod and search) */
     if (ps->lmath == NULL
@@ -164,7 +238,17 @@ ps_reinit(ps_decoder_t *ps, cmd_ln_t *config)
         && cmd_ln_boolean_r(ps->config, "-fwdtree"))
         acmod_set_grow(ps->acmod, TRUE);
 
+    if ((ps->pl_window = cmd_ln_int32_r(ps->config, "-pl_window"))) {
+        /* Initialize an auxiliary phone loop search, which will run in
+         * "parallel" with FSG or N-Gram search. */
+        if ((ps->phone_loop = phone_loop_search_init(ps->config,
+                                                     ps->acmod, ps->dict)) == NULL)
+            return -1;
+        ps->searches = glist_add_ptr(ps->searches, ps->phone_loop);
+    }
+
     /* Dictionary and triphone mappings (depends on acmod). */
+    /* FIXME: pass config, change arguments, implement LTS, etc. */
     if ((ps->dict = dict_init(ps->config, ps->acmod->mdef)) == NULL)
         return -1;
 
@@ -172,8 +256,11 @@ ps_reinit(ps_decoder_t *ps, cmd_ln_t *config)
     if (cmd_ln_str_r(ps->config, "-fsg") || cmd_ln_str_r(ps->config, "-jsgf")) {
         ps_search_t *fsgs;
 
-        if ((fsgs = fsg_search_init(ps->config, ps->acmod, ps->dict)) == NULL)
+        if ((ps->d2p = dict2pid_build(ps->acmod->mdef, ps->dict)) == NULL)
             return -1;
+        if ((fsgs = fsg_search_init(ps->config, ps->acmod, ps->dict, ps->d2p)) == NULL)
+            return -1;
+        fsgs->pls = ps->phone_loop;
         ps->searches = glist_add_ptr(ps->searches, fsgs);
         ps->search = fsgs;
     }
@@ -181,13 +268,20 @@ ps_reinit(ps_decoder_t *ps, cmd_ln_t *config)
              || (lmctl = cmd_ln_str_r(ps->config, "-lmctl"))) {
         ps_search_t *ngs;
 
-        if ((ngs = ngram_search_init(ps->config, ps->acmod, ps->dict)) == NULL)
+        if ((ps->d2p = dict2pid_build(ps->acmod->mdef, ps->dict)) == NULL)
             return -1;
+        if ((ngs = ngram_search_init(ps->config, ps->acmod, ps->dict, ps->d2p)) == NULL)
+            return -1;
+        ngs->pls = ps->phone_loop;
         ps->searches = glist_add_ptr(ps->searches, ngs);
         ps->search = ngs;
     }
     /* Otherwise, we will initialize the search whenever the user
      * decides to load an FSG or a language model. */
+    else {
+        if ((ps->d2p = dict2pid_build(ps->acmod->mdef, ps->dict)) == NULL)
+            return -1;
+    }
 
     /* Initialize performance timer. */
     ps->perf.name = "decode";
@@ -236,6 +330,7 @@ ps_free(ps_decoder_t *ps)
         ps_search_free(gnode_ptr(gn));
     glist_free(ps->searches);
     dict_free(ps->dict);
+    dict2pid_free(ps->d2p);
     acmod_free(ps->acmod);
     logmath_free(ps->lmath);
     cmd_ln_free_r(ps->config);
@@ -274,6 +369,12 @@ ps_get_feat(ps_decoder_t *ps)
     return ps->acmod->fcb;
 }
 
+ps_mllr_t *
+ps_update_mllr(ps_decoder_t *ps, ps_mllr_t *mllr)
+{
+    return acmod_update_mllr(ps->acmod, mllr);
+}
+
 ngram_model_t *
 ps_get_lmset(ps_decoder_t *ps)
 {
@@ -287,32 +388,30 @@ ngram_model_t *
 ps_update_lmset(ps_decoder_t *ps, ngram_model_t *lmset)
 {
     ngram_search_t *ngs;
-    gnode_t *gn;
+    ps_search_t *search;
 
     /* Look for N-Gram search. */
-    for (gn = ps->searches; gn; gn = gnode_next(gn)) {
-        if (0 == strcmp(ps_search_name(gnode_ptr(gn)), "ngram"))
-            break;
-    }
-    if (gn == NULL) {
+    search = ps_find_search(ps, "ngram");
+    if (search == NULL) {
         /* Initialize N-Gram search. */
-        ngs = (ngram_search_t *)ngram_search_init(ps->config,
-                                                  ps->acmod, ps->dict);
-        if (ngs == NULL)
+        search = ngram_search_init(ps->config, ps->acmod, ps->dict, ps->d2p);
+        if (search == NULL)
             return NULL;
-        ps->searches = glist_add_ptr(ps->searches, ngs);
+        search->pls = ps->phone_loop;
+        ps->searches = glist_add_ptr(ps->searches, search);
+        ngs = (ngram_search_t *)search;
     }
     else {
-        ngs = gnode_ptr(gn);
+        ngs = (ngram_search_t *)search;
         /* Free any previous lmset if this is a new one. */
         if (ngs->lmset != NULL && ngs->lmset != lmset)
             ngram_model_free(ngs->lmset);
         ngs->lmset = lmset;
         /* Tell N-Gram search to update its view of the world. */
-        if (ps_search_reinit(ps_search_base(ngs)) < 0)
+        if (ps_search_reinit(search, ps->dict, ps->d2p) < 0)
             return NULL;
     }
-    ps->search = ps_search_base(ngs);
+    ps->search = search;
     return ngs->lmset;
 }
 
@@ -328,28 +427,86 @@ ps_get_fsgset(ps_decoder_t *ps)
 fsg_set_t *
 ps_update_fsgset(ps_decoder_t *ps)
 {
-    gnode_t *gn;
-    fsg_search_t *fsgs;
+    ps_search_t *search;
 
     /* Look for FSG search. */
-    for (gn = ps->searches; gn; gn = gnode_next(gn)) {
-        if (0 == strcmp(ps_search_name(gnode_ptr(gn)), "fsg"))
-            break;
-    }
-    if (gn == NULL) {
+    search = ps_find_search(ps, "fsg");
+    if (search == NULL) {
         /* Initialize FSG search. */
-        fsgs = (fsg_search_t *)fsg_search_init(ps->config,
-                                               ps->acmod, ps->dict);
-        ps->searches = glist_add_ptr(ps->searches, fsgs);
+        search = fsg_search_init(ps->config,
+                                 ps->acmod, ps->dict, ps->d2p);
+        search->pls = ps->phone_loop;
+        ps->searches = glist_add_ptr(ps->searches, search);
     }
     else {
         /* Tell FSG search to update its view of the world. */
-        fsgs = gnode_ptr(gn);
-        if (ps_search_reinit(ps_search_base(fsgs)) < 0)
+        if (ps_search_reinit(search, ps->dict, ps->d2p) < 0)
             return NULL;
     }
-    ps->search = ps_search_base(fsgs);
-    return (fsg_set_t *)fsgs;
+    ps->search = search;
+    return (fsg_set_t *)search;
+}
+
+int
+ps_load_dict(ps_decoder_t *ps, char const *dictfile,
+             char const *fdictfile, char const *format)
+{
+    cmd_ln_t *newconfig;
+    dict2pid_t *d2p;
+    dict_t *dict;
+    gnode_t *gn;
+    int rv;
+
+    /* Create a new scratch config to load this dict (so existing one
+     * won't be affected if it fails) */
+    newconfig = cmd_ln_init(NULL, ps_args(), TRUE, NULL);
+    cmd_ln_set_boolean_r(newconfig, "-dictcase",
+                         cmd_ln_boolean_r(ps->config, "-dictcase"));
+    cmd_ln_set_str_r(newconfig, "-dict", dictfile);
+    if (fdictfile)
+        cmd_ln_set_str_r(newconfig, "-fdict", fdictfile);
+    else
+        cmd_ln_set_str_r(newconfig, "-fdict",
+                         cmd_ln_str_r(ps->config, "-fdict"));
+
+    /* Try to load it. */
+    if ((dict = dict_init(newconfig, ps->acmod->mdef)) == NULL) {
+        cmd_ln_free_r(newconfig);
+        return -1;
+    }
+
+    /* Reinit the dict2pid. */
+    if ((d2p = dict2pid_build(ps->acmod->mdef, dict)) == NULL) {
+        cmd_ln_free_r(newconfig);
+        return -1;
+    }
+
+    /* Success!  Update the existing config to reflect new dicts and
+     * drop everything into place. */
+    cmd_ln_free_r(newconfig);
+    cmd_ln_set_str_r(ps->config, "-dict", dictfile);
+    if (fdictfile)
+        cmd_ln_set_str_r(ps->config, "-fdict", fdictfile);
+    dict_free(ps->dict);
+    ps->dict = dict;
+    dict2pid_free(ps->d2p);
+    ps->d2p = d2p;
+
+    /* And tell all searches to reconfigure themselves. */
+    for (gn = ps->searches; gn; gn = gnode_next(gn)) {
+        ps_search_t *search = gnode_ptr(gn);
+        if ((rv = ps_search_reinit(search, dict, d2p)) < 0)
+            return rv;
+    }
+
+    return 0;
+}
+
+int
+ps_save_dict(ps_decoder_t *ps, char const *dictfile,
+             char const *format)
+{
+    return dict_write(ps->dict, dictfile, format);
 }
 
 int
@@ -360,22 +517,43 @@ ps_add_word(ps_decoder_t *ps,
 {
     int32 wid, lmwid;
     ngram_model_t *lmset;
-    char *pron;
-    int rv;
+    s3cipid_t *pron;
+    char **phonestr, *tmp;
+    int np, i, rv;
 
-    pron = ckd_salloc(phones);
-    if ((wid = dict_add_word(ps->dict, word, pron)) == -1) {
+    /* Parse phones into an array of phone IDs. */
+    tmp = ckd_salloc(phones);
+    np = str2words(tmp, NULL, 0);
+    phonestr = ckd_calloc(np, sizeof(*phonestr));
+    str2words(tmp, phonestr, np);
+    pron = ckd_calloc(np, sizeof(*pron));
+    for (i = 0; i < np; ++i) {
+        pron[i] = bin_mdef_ciphone_id(ps->acmod->mdef, phonestr[i]);
+        if (pron[i] == -1) {
+            E_ERROR("Unknown phone %s in phone string %s\n",
+                    phonestr[i], tmp);
+            ckd_free(phonestr);
+            ckd_free(tmp);
+            ckd_free(pron);
+            return -1;
+        }
+    }
+    /* No longer needed. */
+    ckd_free(phonestr);
+    ckd_free(tmp);
+
+    /* Add it to the dictionary. */
+    if ((wid = dict_add_word(ps->dict, word, pron, np)) == -1) {
         ckd_free(pron);
         return -1;
     }
+    /* No longer needed. */
     ckd_free(pron);
 
+    /* Now we also have to add it to dict2pid. */
+    dict2pid_add_word(ps->d2p, wid);
+
     if ((lmset = ps_get_lmset(ps)) != NULL) {
-        /* FIXME: There is a way more efficient way to do this, since all
-         * we did was replace a placeholder string with the new word
-         * string - therefore what we ought to do is add it directly to
-         * the current LM, then update the mapping without reallocating
-         * everything. */
         /* Add it to the LM set (meaning, the current LM).  In a perfect
          * world, this would result in the same WID, but because of the
          * weird way that word IDs are handled, it doesn't. */
@@ -386,7 +564,7 @@ ps_add_word(ps_decoder_t *ps,
  
     /* Rebuild the widmap and search tree if requested. */
     if (update) {
-        if ((rv = ps_search_reinit(ps->search) < 0))
+        if ((rv = ps_search_reinit(ps->search, ps->dict, ps->d2p) < 0))
             return rv;
     }
     return wid;
@@ -496,9 +674,35 @@ ps_start_utt(ps_decoder_t *ps, char const *uttid)
         acmod_set_rawfh(ps->acmod, rawfh);
     }
 
+    /* Start auxiliary phone loop search. */
+    if (ps->phone_loop)
+        ps_search_start(ps->phone_loop);
+
     return ps_search_start(ps->search);
 }
 
+static int
+ps_search_forward(ps_decoder_t *ps)
+{
+    int nfr;
+
+    nfr = 0;
+    while (ps->acmod->n_feat_frame > 0) {
+        int k;
+        if (ps->phone_loop)
+            if ((k = ps_search_step(ps->phone_loop, ps->acmod->output_frame)) < 0)
+                return k;
+        if (ps->acmod->output_frame >= ps->pl_window)
+            if ((k = ps_search_step(ps->search,
+                                    ps->acmod->output_frame - ps->pl_window)) < 0)
+                return k;
+        acmod_advance(ps->acmod);
+        ++ps->n_frame;
+        ++nfr;
+    }
+    return nfr;
+}
+
 int
 ps_process_raw(ps_decoder_t *ps,
                int16 const *data,
@@ -520,16 +724,13 @@ ps_process_raw(ps_decoder_t *ps,
             return nfr;
 
         /* Score and search as much data as possible */
-        if (!no_search) {
-            while ((nfr = ps_search_step(ps->search)) > 0) {
-                n_searchfr += nfr;
-            }
-            if (nfr < 0)
-                return nfr;
-        }
+        if (no_search)
+            continue;
+        if ((nfr = ps_search_forward(ps)) < 0)
+            return nfr;
+        n_searchfr += nfr;
     }
 
-    ps->n_frame += n_searchfr;
     return n_searchfr;
 }
 
@@ -554,32 +755,44 @@ ps_process_cep(ps_decoder_t *ps,
             return nfr;
 
         /* Score and search as much data as possible */
-        if (!no_search) {
-            while ((nfr = ps_search_step(ps->search)) > 0) {
-                n_searchfr += nfr;
-            }
-            if (nfr < 0)
-                return nfr;
-        }
+        if (no_search)
+            continue;
+        if ((nfr = ps_search_forward(ps)) < 0)
+            return nfr;
+        n_searchfr += nfr;
     }
 
-    ps->n_frame += n_searchfr;
     return n_searchfr;
 }
 
 int
 ps_end_utt(ps_decoder_t *ps)
 {
-    int rv;
+    int rv, i;
 
     acmod_end_utt(ps->acmod);
-    while ((rv = ps_search_step(ps->search)) > 0) {
+
+    /* Search any remaining frames. */
+    if ((rv = ps_search_forward(ps)) < 0) {
+        ptmr_stop(&ps->perf);
+        return rv;
     }
-    if (rv < 0) {
+    /* Finish phone loop search. */
+    if (ps->phone_loop) {
+        if ((rv = ps_search_finish(ps->phone_loop)) < 0) {
+            ptmr_stop(&ps->perf);
+            return rv;
+        }
+    }
+    /* Search any frames remaining in the lookahead window. */
+    for (i = ps->acmod->output_frame - ps->pl_window;
+         i < ps->acmod->output_frame; ++i)
+        ps_search_step(ps->search, i);
+    /* Finish main search. */
+    if ((rv = ps_search_finish(ps->search)) < 0) {
         ptmr_stop(&ps->perf);
         return rv;
     }
-    rv = ps_search_finish(ps->search);
     ptmr_stop(&ps->perf);
 
     /* Log a backtrace if requested. */
@@ -712,8 +925,8 @@ ps_nbest(ps_decoder_t *ps, int sf, int ef,
         lwf = ((ngram_search_t *)ps->search)->bestpath_fwdtree_lw_ratio;
     }
 
-    w1 = ctx1 ? dict_to_id(ps_search_dict(ps->search), ctx1) : -1;
-    w2 = ctx2 ? dict_to_id(ps_search_dict(ps->search), ctx2) : -1;
+    w1 = ctx1 ? dict_wordid(ps_search_dict(ps->search), ctx1) : -1;
+    w2 = ctx2 ? dict_wordid(ps_search_dict(ps->search), ctx2) : -1;
     nbest = ps_astar_start(dag, lmset, lwf, sf, ef, w1, w2);
 
     return (ps_nbest_t *)nbest;
@@ -741,19 +954,19 @@ ps_nbest_next(ps_nbest_t *nbest)
 char const *
 ps_nbest_hyp(ps_nbest_t *nbest, int32 *out_score)
 {
-    if (nbest->paths_done == NULL)
+    if (nbest->top == NULL)
         return NULL;
-    if (out_score) *out_score = nbest->paths_done->score;
-    return ps_astar_hyp(nbest, nbest->paths_done);
+    if (out_score) *out_score = nbest->top->score;
+    return ps_astar_hyp(nbest, nbest->top);
 }
 
 ps_seg_t *
 ps_nbest_seg(ps_nbest_t *nbest, int32 *out_score)
 {
-    if (nbest->paths_done == NULL)
+    if (nbest->top == NULL)
         return NULL;
-    if (out_score) *out_score = nbest->paths_done->score;
-    return ps_astar_seg_iter(nbest, nbest->paths_done, 1.0);
+    if (out_score) *out_score = nbest->top->score;
+    return ps_astar_seg_iter(nbest, nbest->top, 1.0);
 }
 
 int
@@ -788,22 +1001,63 @@ ps_get_all_time(ps_decoder_t *ps, double *out_nspeech,
 
 void
 ps_search_init(ps_search_t *search, ps_searchfuncs_t *vt,
-               cmd_ln_t *config, acmod_t *acmod, dict_t *dict)
+               cmd_ln_t *config, acmod_t *acmod, dict_t *dict,
+               dict2pid_t *d2p)
 {
     search->vt = vt;
     search->config = config;
     search->acmod = acmod;
-    search->dict = dict;
-    search->start_wid = dict_to_id(dict, "<s>");
-    search->finish_wid = dict_to_id(dict, "</s>");
-    search->silence_wid = dict_to_id(dict, "<sil>");
+    if (d2p)
+        search->d2p = dict2pid_retain(d2p);
+    else
+        search->d2p = NULL;
+    if (dict) {
+        search->dict = dict_retain(dict);
+        search->start_wid = dict_startwid(dict);
+        search->finish_wid = dict_finishwid(dict);
+        search->silence_wid = dict_silwid(dict);
+        search->n_words = dict_size(dict);
+    }
+    else {
+        search->dict = NULL;
+        search->start_wid = search->finish_wid = search->silence_wid = -1;
+        search->n_words = 0;
+    }
 }
 
+void
+ps_search_base_reinit(ps_search_t *search, dict_t *dict,
+                      dict2pid_t *d2p)
+{
+    dict_free(search->dict);
+    dict2pid_free(search->d2p);
+    /* FIXME: _retain() should just return NULL if passed NULL. */
+    if (dict) {
+        search->dict = dict_retain(dict);
+        search->start_wid = dict_startwid(dict);
+        search->finish_wid = dict_finishwid(dict);
+        search->silence_wid = dict_silwid(dict);
+        search->n_words = dict_size(dict);
+    }
+    else {
+        search->dict = NULL;
+        search->start_wid = search->finish_wid = search->silence_wid = -1;
+        search->n_words = 0;
+    }
+    if (d2p)
+        search->d2p = dict2pid_retain(d2p);
+    else
+        search->d2p = NULL;
+}
+
+
 void
 ps_search_deinit(ps_search_t *search)
 {
     /* FIXME: We will have refcounting on acmod, config, etc, at which
      * point we will free them here too. */
+    dict_free(search->dict);
+    dict2pid_free(search->d2p);
     ckd_free(search->hyp_str);
     ps_lattice_free(search->dag);
 }
