@@ -42,11 +42,21 @@
 #include <pio.h>
 #include <err.h>
 #include <strfuncs.h>
+#include <filename.h>
 #include <byteorder.h>
 
 /* PocketSphinx headers. */
 #include <pocketsphinx.h>
 
+/* S3kr3t headerz. */
+#include "pocketsphinx_internal.h"
+
+/* Silvio Moioli: setbuf doesn't exist in Windows CE */
+#if defined(_WIN32_WCE)
+	void setbuf(FILE* file, char* buf){
+	}
+#endif
+
 static const arg_t ps_args_def[] = {
     POCKETSPHINX_OPTIONS,
     /* Various options specific to batch-mode processing. */
@@ -72,6 +82,34 @@ static const arg_t ps_args_def[] = {
       ARG_INT32,
       "1",
       "Do every Nth line in the control file" },
+    { "-mllrctl",
+      ARG_STRING,
+      NULL,
+      "Control file listing MLLR transforms to use for each utterance" },
+    { "-mllrdir",
+      ARG_STRING,
+      NULL,
+      "Base directory for MLLR transforms" },
+    { "-mllrext",
+      ARG_STRING,
+      NULL,
+      "File extension for MLLR transforms (including leading dot)" },
+    { "-lmnamectl",
+      ARG_STRING,
+      NULL,
+      "Control file listing LM name to use for each utterance" },
+    { "-fsgctl",
+      ARG_STRING,
+      NULL,
+      "Control file listing FSG file to use for each utterance" },
+    { "-fsgdir",
+      ARG_STRING,
+      NULL,
+      "Base directory for FSG files" },
+    { "-fsgext",
+      ARG_STRING,
+      NULL,
+      "File extension for FSG files (including leading dot)" },
 
     /* Input file types and locations. */
     { "-adcin",
@@ -96,18 +134,22 @@ static const arg_t ps_args_def[] = {
       ARG_STRING,
       NULL,
       "Recognition output file name" },
-    { "-hypconf",
-      ARG_STRING,
-      NULL,
-      "Recognition output with confidence file name" },
     { "-hypseg",
       ARG_STRING,
       NULL,
       "Recognition output with segmentation file name" },
+    { "-ctm",
+      ARG_STRING,
+      NULL,
+      "Recognition output in CTM file format (may require post-sorting)" },
     { "-outlatdir",
       ARG_STRING,
       NULL,
       "Directory for dumping word lattices" },
+    { "-build_outdirs",
+      ARG_BOOLEAN,
+      "yes",
+      "Create missing subdirectories in output directory" },
     { "-nbestdir",
       ARG_STRING,
       NULL,
@@ -159,7 +201,12 @@ read_mfc_file(FILE *infh, int sf, int ef, int *out_nfr, int ceplen)
     nfr = ef - sf;
     mfcs = ckd_calloc_2d(nfr, ceplen, sizeof(**mfcs));
     floats = (float32 *)mfcs[0];
-    fread(floats, 4, nfr * ceplen, infh);
+    if (fread(floats, 4, nfr * ceplen, infh) != nfr * ceplen) {
+        E_ERROR_SYSTEM("Failed to read %d items from mfcfile");
+        fclose(infh);
+        ckd_free_2d(mfcs);
+        return NULL;
+    }
     if (swap) {
         for (i = 0; i < nfr * ceplen; ++i)
             SWAP_FLOAT32(&floats[i]);
@@ -172,6 +219,106 @@ read_mfc_file(FILE *infh, int sf, int ef, int *out_nfr, int ceplen)
     return mfcs;
 }
 
+static int
+process_mllrctl_line(ps_decoder_t *ps, cmd_ln_t *config, char const *file)
+{
+    char const *mllrdir, *mllrext;
+    char *infile = NULL;
+    ps_mllr_t *mllr;
+    static char *lastfile;
+
+    if (file == NULL)
+        return 0;
+
+    if (lastfile && 0 == strcmp(file, lastfile))
+        return 0;
+
+    ckd_free(lastfile);
+    lastfile = ckd_salloc(file);
+
+    mllrext = cmd_ln_str_r(config, "-mllrext");
+    if ((mllrdir = cmd_ln_str_r(config, "-mllrdir")))
+        infile = string_join(mllrdir, "/", file, 
+                             mllrext ? mllrext : "", NULL);
+    else if (mllrext)
+        infile = string_join(file, mllrext, NULL);
+    else
+        infile = ckd_salloc(file);
+
+    if ((mllr = ps_mllr_read(infile)) == NULL) {
+        ckd_free(infile);
+        return -1;
+    }
+    if (ps_update_mllr(ps, mllr) == NULL) {
+        ps_mllr_free(mllr);
+        ckd_free(infile);
+        return -1;
+    }
+
+    ckd_free(infile);
+    return 0;
+}
+
+static int
+process_fsgctl_line(ps_decoder_t *ps, cmd_ln_t *config, char const *file)
+{
+    fsg_set_t *fsgset = ps_get_fsgset(ps);
+    fsg_model_t *fsg;
+    char const *fsgdir, *fsgext;
+    char *infile = NULL;
+    static char *lastfile;
+
+    if (file == NULL)
+        return 0;
+
+    if (lastfile && 0 == strcmp(file, lastfile))
+        return 0;
+
+    fsgext = cmd_ln_str_r(config, "-fsgext");
+    if ((fsgdir = cmd_ln_str_r(config, "-fsgdir")))
+        infile = string_join(fsgdir, "/", file,
+                             fsgext ? fsgext : "", NULL);
+    else if (fsgext)
+        infile = string_join(file, fsgext, NULL);
+    else
+        infile = ckd_salloc(file);
+
+    if ((fsg = fsg_model_readfile(infile, ps_get_logmath(ps),
+                                  cmd_ln_float32_r(config, "-lw"))) == NULL)
+        goto error_out;
+
+    if (fsgset == NULL)
+        fsgset = ps_update_fsgset(ps);
+    if (lastfile)
+        fsg_set_remove_byname(fsgset, lastfile);
+
+    ckd_free(lastfile);
+    lastfile = ckd_salloc(file);
+
+    fsg_set_add(fsgset, lastfile, fsg);
+    fsg_set_select(fsgset, lastfile);
+
+    ps_update_fsgset(ps);
+error_out:
+    ckd_free(infile);
+    return 0;
+}
+
+static int
+process_lmnamectl_line(ps_decoder_t *ps, cmd_ln_t *config, char const *lmname)
+{
+    ngram_model_t *lmset = ps_get_lmset(ps);
+
+    if (lmname == NULL)
+        return 0;
+    if (ngram_model_set_select(lmset, lmname) == NULL) {
+        E_ERROR("No such language model: %s\n", lmname);
+        return -1;
+    }
+    ps_update_lmset(ps, lmset);
+    return 0;
+}
+
 static int
 process_ctl_line(ps_decoder_t *ps, cmd_ln_t *config,
                  char const *file, char const *uttid, int32 sf, int32 ef)
@@ -180,6 +327,11 @@ process_ctl_line(ps_decoder_t *ps, cmd_ln_t *config,
     char const *cepdir, *cepext;
     char *infile;
 
+    if (ef != -1 && ef < sf) {
+        E_ERROR("End frame %d is < start frame %d\n", ef, sf);
+        return -1;
+    }
+    
     cepdir = cmd_ln_str_r(config, "-cepdir");
     cepext = cmd_ln_str_r(config, "-cepext");
 
@@ -195,18 +347,18 @@ process_ctl_line(ps_decoder_t *ps, cmd_ln_t *config,
         return -1;
     }
     if (cmd_ln_boolean_r(config, "-adcin")) {
-        fseek(infh, cmd_ln_int32_r(config, "-adchdr"), SEEK_SET);
+        
         if (ef != -1) {
             ef = (int32)((ef - sf)
                          * (cmd_ln_float32_r(config, "-samprate")
                             / cmd_ln_int32_r(config, "-frate"))
                          + (cmd_ln_float32_r(config, "-samprate")
-                            * cmd_ln_int32_r(config, "-wlen")));
+                            * cmd_ln_float32_r(config, "-wlen")));
         }
         sf = (int32)(sf
                      * (cmd_ln_float32_r(config, "-samprate")
                         / cmd_ln_int32_r(config, "-frate")));
-        fseek(infh, sf * sizeof(int16), SEEK_CUR);
+        fseek(infh, cmd_ln_int32_r(config, "-adchdr") + sf * sizeof(int16), SEEK_SET);
         ps_decode_raw(ps, infh, uttid, ef);
     }
     else {
@@ -240,6 +392,14 @@ write_lattice(ps_decoder_t *ps, char const *latdir, char const *uttid)
         return -1;
     }
     outfile = string_join(latdir, "/", uttid, ".lat", NULL);
+    /* Build output directory structure if possible/requested (it is
+     * by default). */
+    if (cmd_ln_boolean_r(ps_get_config(ps), "-build_outdirs")) {
+        char *dirname = ckd_salloc(outfile);
+        path2dirname(outfile, dirname);
+        build_directory(dirname);
+        ckd_free(dirname);
+    }
     if (ps_lattice_write(lat, outfile) < 0) {
         E_ERROR("Failed to write lattice to %s\n", outfile);
         return -1;
@@ -286,18 +446,55 @@ write_hypseg(FILE *fh, ps_decoder_t *ps, char const *uttid)
 }
 
 static int
-write_hypconf(FILE *fh, ps_decoder_t *ps, ps_seg_t *itor, char const *uttid, int32 score)
+write_ctm(FILE *fh, ps_decoder_t *ps, ps_seg_t *itor, char const *uttid, int32 frate)
 {
     logmath_t *lmath = ps_get_logmath(ps);
+    char *dupid, *show, *channel, *c;
+    double ustart = 0.0;
+
+    /* We have semi-standardized on comma-separated uttids which
+     * correspond to the fields of the STM file.  So if there's a
+     * comma in the uttid, take the first two fields as show and
+     * channel, and also try to find the start time. */
+    show = dupid = ckd_salloc(uttid);
+    if ((c = strchr(dupid, ',')) != NULL) {
+        *c++ = '\0';
+        channel = c;
+        if ((c = strchr(c, ',')) != NULL) {
+            *c++ = '\0';
+            if ((c = strchr(c, ',')) != NULL) {
+                ustart = atof_c(c + 1);
+            }
+        }
+    }
+    else {
+        channel = NULL;
+    }
 
     while (itor) {
-        char const *w = ps_seg_word(itor);
-        int32 prob = ps_seg_prob(itor, NULL, NULL, NULL);
+        int32 prob, sf, ef, wid;
+        char const *w;
 
-        fprintf(fh, "%s(%.2f) ", w, logmath_log_to_ln(lmath, prob));
+        /* Skip things that aren't "real words" (FIXME: currently
+         * requires s3kr3t h34d3rz...) */
+        w = ps_seg_word(itor);
+        wid = dict_wordid(ps->dict, w);
+        if (wid >= 0 && dict_real_word(ps->dict, wid)) {
+            prob = ps_seg_prob(itor, NULL, NULL, NULL);
+            ps_seg_frames(itor, &sf, &ef);
+        
+            fprintf(fh, "%s %s %.2f %.2f %s %.3f\n",
+                    show,
+                    channel ? channel : "1",
+                    ustart + (double)sf / frate,
+                    (double)(ef - sf) / frate,
+                    /* FIXME: More s3kr3tz */
+                    dict_basestr(ps->dict, wid),
+                    logmath_exp(lmath, prob));
+        }
         itor = ps_seg_next(itor);
     }
-    fprintf(fh, "(%s %.2f)\n", uttid, logmath_log_to_ln(lmath, ps_get_prob(ps, NULL)));
+    ckd_free(dupid);
 
     return 0;
 }
@@ -309,57 +506,111 @@ process_ctl(ps_decoder_t *ps, cmd_ln_t *config, FILE *ctlfh)
     int32 i;
     char *line;
     size_t len;
-    FILE *hypfh = NULL, *hypsegfh = NULL, *hypconffh = NULL;
+    FILE *hypfh = NULL, *hypsegfh = NULL, *ctmfh = NULL;
+    FILE *mllrfh = NULL, *lmfh = NULL, *fsgfh = NULL;
     double n_speech, n_cpu, n_wall;
     char const *outlatdir;
     char const *nbestdir;
+    char const *str;
+    int frate;
 
     ctloffset = cmd_ln_int32_r(config, "-ctloffset");
     ctlcount = cmd_ln_int32_r(config, "-ctlcount");
     ctlincr = cmd_ln_int32_r(config, "-ctlincr");
     outlatdir = cmd_ln_str_r(config, "-outlatdir");
     nbestdir = cmd_ln_str_r(config, "-nbestdir");
+    frate = cmd_ln_int32_r(config, "-frate");
 
-    if (cmd_ln_str_r(config, "-hyp")) {
-        hypfh = fopen(cmd_ln_str_r(config, "-hyp"), "w");
+    if ((str = cmd_ln_str_r(config, "-mllrctl"))) {
+        mllrfh = fopen(str, "r");
+        if (mllrfh == NULL) {
+            E_ERROR_SYSTEM("Failed to open MLLR control file file %s", str);
+            goto done;
+        }
+    }
+    if ((str = cmd_ln_str_r(config, "-fsgctl"))) {
+        fsgfh = fopen(str, "r");
+        if (fsgfh == NULL) {
+            E_ERROR_SYSTEM("Failed to open FSG control file file %s", str);
+            goto done;
+        }
+    }
+    if ((str = cmd_ln_str_r(config, "-lmnamectl"))) {
+        lmfh = fopen(str, "r");
+        if (lmfh == NULL) {
+            E_ERROR_SYSTEM("Failed to open LM name control file file %s", str);
+            goto done;
+        }
+    }
+    if ((str = cmd_ln_str_r(config, "-hyp"))) {
+        hypfh = fopen(str, "w");
         if (hypfh == NULL) {
-            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing", hypfh);
-            return;
+            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing", str);
+            goto done;
         }
         setbuf(hypfh, NULL);
     }
-    if (cmd_ln_str_r(config, "-hypseg")) {
-        hypsegfh = fopen(cmd_ln_str_r(config, "-hypseg"), "w");
+    if ((str = cmd_ln_str_r(config, "-hypseg"))) {
+        hypsegfh = fopen(str, "w");
         if (hypsegfh == NULL) {
-            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing",
-                           hypsegfh);
-            return;
+            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing", str);
+            goto done;
         }
         setbuf(hypsegfh, NULL);
     }
-    if (cmd_ln_str_r(config, "-hypconf")) {
-        hypconffh = fopen(cmd_ln_str_r(config, "-hypconf"), "w");
-        if (hypconffh == NULL) {
-            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing",
-                           hypconffh);
-            return;
+    if ((str = cmd_ln_str_r(config, "-ctm"))) {
+        ctmfh = fopen(str, "w");
+        if (ctmfh == NULL) {
+            E_ERROR_SYSTEM("Failed to open hypothesis file %s for writing", str);
+            goto done;
         }
-        setbuf(hypconffh, NULL);
+        setbuf(ctmfh, NULL);
     }
 
     i = 0;
     while ((line = fread_line(ctlfh, &len))) {
         char *wptr[4];
         int32 nf, sf, ef;
+        char *mllrline = NULL, *lmline = NULL, *fsgline = NULL;
+        char *fsgfile = NULL, *lmname = NULL, *mllrfile = NULL;
+
+        if (mllrfh) {
+            mllrline = fread_line(mllrfh, &len);
+            if (mllrline == NULL) {
+                E_ERROR("File size mismatch between control and MLLR control\n");
+                ckd_free(line);
+                ckd_free(mllrline);
+                goto done;
+            }
+            mllrfile = string_trim(mllrline, STRING_BOTH);
+        }
+        if (lmfh) {
+            lmline = fread_line(lmfh, &len);
+            if (lmline == NULL) {
+                E_ERROR("File size mismatch between control and LM control\n");
+                ckd_free(line);
+                ckd_free(lmline);
+                goto done;
+            }
+            lmname = string_trim(lmname, STRING_BOTH);
+        }
+        if (fsgfh) {
+            fsgline = fread_line(fsgfh, &len);
+            if (fsgline == NULL) {
+                E_ERROR("File size mismatch between control and FSG control\n");
+                ckd_free(line);
+                ckd_free(fsgline);
+                goto done;
+            }
+            fsgfile = string_trim(fsgline, STRING_BOTH);
+        }
 
         if (i < ctloffset) {
             i += ctlincr;
-            ckd_free(line);
-            continue;
+            goto nextline;
         }
         if (ctlcount != -1 && i >= ctloffset + ctlcount) {
-            ckd_free(line);
-            break;
+            goto nextline;
         }
 
         sf = 0;
@@ -384,19 +635,22 @@ process_ctl(ps_decoder_t *ps, cmd_ln_t *config, FILE *ctlfh)
             if (nf > 3)
                 uttid = wptr[3];
             /* Do actual decoding. */
+            process_mllrctl_line(ps, config, mllrfile);
+            process_lmnamectl_line(ps, config, lmname);
+            process_fsgctl_line(ps, config, fsgfile);
             process_ctl_line(ps, config, file, uttid, sf, ef);
             hyp = ps_get_hyp(ps, &score, &uttid);
             
             /* Write out results and such. */
             if (hypfh) {
-                fprintf(hypfh, "%s (%s %d)\n", hyp, uttid, score);
+                fprintf(hypfh, "%s (%s %d)\n", hyp ? hyp : "", uttid, score);
             }
             if (hypsegfh) {
                 write_hypseg(hypsegfh, ps, uttid);
             }
-            if (hypconffh) {
+            if (ctmfh) {
                 ps_seg_t *itor = ps_seg_iter(ps, &score);
-                write_hypconf(hypconffh, ps, itor, uttid, score);
+                write_ctm(ctmfh, ps, itor, uttid, frate);
             }
             if (outlatdir) {
                 write_lattice(ps, outlatdir, uttid);
@@ -409,8 +663,12 @@ process_ctl(ps_decoder_t *ps, cmd_ln_t *config, FILE *ctlfh)
             E_INFO("%s: %.2f xRT (CPU), %.2f xRT (elapsed)\n",
                    uttid, n_cpu / n_speech, n_wall / n_speech);
         }
-        ckd_free(line);
         i += ctlincr;
+    nextline:
+        ckd_free(mllrline);
+        ckd_free(fsgline);
+        ckd_free(lmline);
+        ckd_free(line);
     }
 
     ps_get_all_time(ps, &n_speech, &n_cpu, &n_wall);
@@ -419,12 +677,13 @@ process_ctl(ps_decoder_t *ps, cmd_ln_t *config, FILE *ctlfh)
     E_INFO("AVERAGE %.2f xRT (CPU), %.2f xRT (elapsed)\n",
            n_cpu / n_speech, n_wall / n_speech);
 
+done:
     if (hypfh)
         fclose(hypfh);
     if (hypsegfh)
         fclose(hypsegfh);
-    if (hypconffh)
-        fclose(hypconffh);
+    if (ctmfh)
+        fclose(ctmfh);
 }
 
 int
@@ -467,3 +726,28 @@ main(int32 argc, char *argv[])
     ps_free(ps);
     return 0;
 }
+
+/** Silvio Moioli: Windows CE/Mobile entry point added. */
+#if defined(_WIN32_WCE)
+#pragma comment(linker,"/entry:mainWCRTStartup")
+#include <windows.h>
+
+//Windows Mobile has the Unicode main only
+int wmain(int32 argc, wchar_t *wargv[]) {
+    char** argv;
+    size_t wlen;
+    size_t len;
+    int i;
+
+    argv = malloc(argc*sizeof(char*));
+    for (i=0; i<argc; i++){
+        wlen = lstrlenW(wargv[i]);
+        len = wcstombs(NULL, wargv[i], wlen);
+        argv[i] = malloc(len+1);
+        wcstombs(argv[i], wargv[i], wlen);
+    }
+
+    //assuming ASCII parameters
+    return main(argc, argv);
+}
+#endif
